#!/usr/bin/env python3

"""Marketa command line interface
Usage:
    marketa (status | watch <symbol> | unwatch <symbol> | pull | clean | monitor | install | uninstall)
    marketa -h|--help
    marketa -v|--version
Commands:
    status              reports it's current configuration status
    watch <symbol>      adds symbol to a watchlist
    unwatch <symbol>    removes symbol from watchlist
    pull                gets latest historical data for symbols in watchlist refreshing local caches
    clean               removes cached data
    monitor             runs marketa in 'monitor' service mode, getting regular updates for watchlist symbols
    install             installs, enables and starts marketa 'monitor' mode as a systemd service
    uninstall           uninstalls marketa systemd service
Options:
    -h --help           show this screen
    -v --version        show version
"""

import os, sys, subprocess
from time import sleep
from pathlib import Path

from docopt import docopt
from exitstatus import ExitStatus
from datetime import datetime
import humanize
from marketa.shared.utils import hookup_tqdm
from marketa.shared.exceptions import MarketaException
from marketa import get_facade
mk = get_facade()


def pull():
    print('pulling data for symbols in the watchlist ...')
    mk.pull()
    print('data for symbols in watchlist was succesfully updated')


def watch(symbol:str):
    if mk.is_watching(symbol):
        raise MarketaException(f'symbol "{symbol}" is already in the watchlist')

    mk.watch(symbol)
    print(f'symbol "{symbol}" added to the watchlist')


def unwatch(symbol:str):
    if not mk.is_watching(symbol):
        raise MarketaException(f'symbol "{symbol}" is not in the watchlist')

    mk.unwatch(symbol)
    print(f'symbol "{symbol}" removed from the watchlist')


def monitor():
    try:
        while True:
            print('marketa is running in monitor mode!', flush=True)
            sleep(5)
        
    except BaseException as e:
        print(f'unexpected error: {e}')
        raise e


src = Path(__file__).parent.joinpath('marketa.service')
dst = '/etc/systemd/system/marketa.service'

def install():
    print('installing service ...')
    subprocess.check_call(f'sudo cp {src} {dst}', shell=True)
    subprocess.check_call('sudo systemctl daemon-reload', shell=True)
    subprocess.check_call('sudo systemctl enable marketa', shell=True)
    print('starting service ...')
    subprocess.check_call('sudo systemctl start marketa', shell=True)
    subprocess.check_call('systemctl status marketa', shell=True)
    print('service installed successfully and running')

    
def uninstall():
    print('stopping service ...')
    subprocess.check_call('sudo systemctl stop marketa', shell=True)
    subprocess.check_call('sudo systemctl disable marketa', shell=True)
    subprocess.check_call(f'sudo rm {dst}', shell=True)
    subprocess.check_call('sudo systemctl daemon-reload', shell=True)
    print('service uninstalled successfully')


def status():
    s = mk.get_state()
    watchlist = s['watchlists']['default']
    result = f'watching {len(watchlist)} symbols: '
    result+= ', '.join(watchlist)
    result+='\n'
    result+=' data last updated: '
    if 'data_updated_on' in s.keys():
        result += humanize.naturaltime(datetime.now() - s['data_updated_on'])
    else:
        result+='never'
    
    print(result)


def clean():
    mk.clear_caches()
    print('cache cleaned succesfully')


if __name__ == '__main__':
    try:
        arguments = docopt(str(__doc__), version='0.1.0') # semver MAJOR.MINOR.PATCH
        hookup_tqdm() # enable tqdm progress bar

        if arguments['pull']:
            pull()
        elif arguments['clean']:
            clean()
        elif arguments['watch']:
            watch(arguments['<symbol>'][0])
        elif arguments['unwatch']:
            unwatch(arguments['<symbol>'][0])
        elif arguments['status']:
            status()
        elif arguments['monitor']:
            monitor()
        elif arguments['install']:
            install()
        elif arguments['uninstall']:
            uninstall()
        else:
            print('invalid arguments specified')

    except MarketaException as e:
        print(f'error: {e}')
        sys.exit(ExitStatus.failure)

    except Exception as e:
        print(f'unexpected error: {e}')
        raise e
        sys.exit(ExitStatus.failure)

    sys.exit(ExitStatus.success)



from typing import Any, Callable, TypeVar

from pandas import DataFrame
from diskcache import Cache


def search(self: DataFrame, val: str, case=False, regex=False):
    """Search all the text columns of `df`, return rows with any matches."""
    textlikes = self.select_dtypes(include=[object, "string"])
    return self[
        textlikes.apply(
            lambda column: column.str.contains(val, regex=regex, case=case, na=False)
        ).any(axis=1)
    ]
DataFrame.search = search
del(search) # clean up namespace


def drop_rows(self: DataFrame, column: str, predicate: Callable[[object], bool]):
    indexNames = self[lambda x: predicate(self[column])].index
    self.drop(indexNames , inplace=True)
DataFrame.drop_rows = drop_rows
del(drop_rows)


T = TypeVar("T")
def getf(self:Cache, key: str, getter: Callable[[], T], timeout:int) -> T:
    val = self.get(key)
    if val is None:
        val = getter()
        self.set(key, val, expire=timeout)
    return val
Cache.getf = getf
del(getf)


from sqlalchemy import create_engine, false
from sqlalchemy.orm import sessionmaker, Session
from .base import Base


class RepositoryBase:
    def __init__(self, in_memory=False):
        
        if in_memory:
            self._engine = create_engine('sqlite://', echo=False)
            self._engine.execute("ATTACH DATABASE ':memory:' AS my_database")
        else:
            #engine = create_engine('postgresql://usr:pass@localhost:5432/sqlalchemy')
            self._engine = create_engine('sqlite:///marketa.sqlite', echo = False)
        
        self._create_session = sessionmaker(bind=self._engine)
        self._session: Session = self._create_session()
    

    def create_all(self):
        Base.metadata.create_all(self._engine)
        self.commit()

    def recreate_all(self):
        Base.metadata.drop_all(self._engine)
        self.commit()
        self.create_all()

    def commit(self):
        self._session.commit()


    def __enter__(self):
        self._ensure_session()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._session.__exit__(exc_type, exc_value, traceback)
        self._session = None


    def _ensure_session(self):
        if not self._session:
            self._session = self._create_session()



from .base import Base
from sqlalchemy import Column, String

class Application(Base):
    """Container for global application settings"""

    __tablename__ = 'application'

    key = Column(String, primary_key=True)
    value = Column(String)



import shutil
import matplotlib
from matplotlib import pyplot as plt
from .services.feed import Feed
from .entities.ticker import Ticker
from .shared.utils import  check_internet as is_online
from .shared.exceptions import MarketaException
from .data.repository import Repository
from .shared.taskeeper import taskeeper

class Facade:

    def __init__(self, 
        repository: Repository, 
        feed_provider: Feed):
        self._repository = repository
        self._feed_provider = feed_provider
        repository.create_all()


    def is_watching(self, symbol:str):
        raise NotImplementedError()
        symbol = symbol.upper()
        state = load_state()
        watchlist = state['watchlists']['default']
        return symbol in watchlist


    def watch(self, symbol:str):
        raise NotImplementedError()
        symbol = symbol.upper()
        state = load_state()
        watchlist = state['watchlists']['default']
        if symbol in watchlist:
            raise Exception(f'symbol {symbol} is already in the watchlist')
        watchlist.append(symbol)
        save_state(state)


    def unwatch(self, symbol:str):
        raise NotImplementedError()
        symbol = symbol.upper()
        state = load_state()
        watchlist = state['watchlists']['default']
        if symbol not in watchlist:
            raise Exception(f'symbol {symbol} is not in the watchlist') 
        watchlist.remove(symbol)    
        save_state(state)


    def pull(self, make_inet_check = True):
        if make_inet_check and not is_online(): 
            raise MarketaException('unable to pull data as there is no internet connection available')

        watchlist = self._repository.get_watchlist_symbols()

        with taskeeper(steps = len(watchlist) + 2) as task:
            with task.step('loading tradable nasdaq symbols'):
                tradable_instruments = self._feed_provider.load_instruments()
            
            with task.step('updating database'):
                self._repository.upsert_instruments(tradable_instruments)
            
            tradable_symbols = list(map(lambda x: x.symbol, tradable_instruments))
            for s in watchlist:
                if s not in tradable_symbols:
                    raise Exception(f'symbol "{s}" in not listed on nasdaq exchange')

                with task.step(f'loading "{s}" symbol data'):
                    data = self._feed_provider.load_instrument_data(s)
                    self._repository.upsert_instrument_data(data)
                    self._repository.update_watchlist_item_timestamp(s)
                
            task.set_info(f'data successfully loaded for {len(watchlist)} symbol(s)')


    def get_state(self):
        raise NotImplementedError()
        state = load_state()
        return state


    def clear_caches(self):
        raise NotImplementedError()
        path = get_state_path().parent.resolve()
        shutil.rmtree(path, ignore_errors=True)


    def get_ticker(self, symbol: str) -> Ticker:
        return self._feed_provider.load_instrument_data(symbol)


    def plot(self, ticker: Ticker):
        mpl_backend = 'QTAgg'
        matplotlib.use(mpl_backend)
        with plt.ion():
            plt.plot([1,2,3])
        raise NotImplementedError()




from datetime import datetime
from decimal import Decimal
from typing import Iterable
from zoneinfo import ZoneInfo

from marketa.data.instrument_data import InstrumentData
from marketa.shared.exceptions import MarketaException
from .repository_base import RepositoryBase

# ensure all entities are imported,
# so that they will be created automatically with repo.create_all() call
from .application import Application
from .market import Market
from .watchlist import Watchlist
from .instrument import Instrument
from .price_daily import PriceDaily
from .price_intraday import PriceIntraday


class Repository(RepositoryBase):
    def __init__(self, in_memory=False):
        super().__init__(in_memory)

    def create_all(self):
        super(Repository, self).create_all()

        # seed database if there are no symbols yet
        if len(self.get_watchlist_symbols()) == 0:
            self.add_watchlist_item('AAAU')


    def create_market(self, market: Market):
        self._session.add(market)

    
    def get_markets(self) -> Iterable[Market]:
        return self._session.query(Market).all()


    def create_instrument(self, instrument: Instrument):
        if instrument.symbol is None:
            raise MarketaException('symbol for the instrument is none')
        self._session.add(instrument)


    def get_instruments(self) -> Iterable[Instrument]:
        return self._session.query(Instrument).all()


    def get_watchlist_symbols(self):
        items = self._session.query(Watchlist).all()
        if len(items) == 0:
            return []
        return list(map(lambda x: x.symbol, items))

    def add_watchlist_item(self, symbol: str):
        item = Watchlist()
        item.symbol = symbol
        item.updated_at = gmt_time_now()
        self._session.add(item)
        self.commit()

    def update_watchlist_item_timestamp(self, symbol: str):
        item: Watchlist = self._session.query(Watchlist).one()
        item.updated_at = gmt_time_now()
        self.commit()


    def create_intraday_price(self, symbol: str, data: PriceIntraday):
        market: Market = self._session.query(Market).one()
        instrument: Instrument = (self._session
            .query(Instrument)
            .filter(Instrument.symbol == symbol)
            .one()
        )
        data.instrument_id = instrument.id
        data.market_id = market.id
        self._session.add(data)


    def get_intraday_prices(self, symbol: str) -> Iterable[Instrument]:
        market: Market = self._session.query(Market).one()
        instrument: Instrument = (self._session
            .query(Instrument)
            .filter(Instrument.symbol == symbol)
            .one()
        )
        prices: PriceIntraday = (self._session
            .query(PriceIntraday)
            .filter(
                PriceIntraday.instrument_id == instrument.id, 
                PriceIntraday.market_id == market.id
            )
            .all()
        )
        return prices
        

    def upsert_instruments(self, instruments):
        for instrument in instruments:
            self.upsert_instrument(instrument, commit=False)
        self.commit()

    def upsert_instrument(self, data: Instrument, commit=True):
        instrument: Instrument = (self._session
            .query(Instrument)
            .filter(
                Instrument.symbol == data.symbol
            )
            .one_or_none()
        )
        if instrument is None:
            self.create_instrument(data)
        else:
            if data.symbol is not None:
                instrument.symbol = data.symbol
            if data.name is not None:
                instrument.name = data.name
            if data.description is not None:
                instrument.description = data.description
            if data.type is not None:
                instrument.type = data.type
            if data.currency is not None:
                instrument.currency = data.currency

        if commit:
            self.commit()

    

    def upsert_instrument_data(self, data: InstrumentData):
        market: Market = self._session.query(Market).one()
        self.upsert_instrument(data.instrument)
        instrument: Instrument = (self._session
            .query(Instrument)
            .filter(
                Instrument.symbol == data.instrument.symbol
            )
            .one()
        )
        
        price = data.price_intraday
        now = gmt_time_now()
        price.datetime = now
        price.imported_at = now
        
        price.instrument_id = instrument.id
        price.market_id = market.id

        self.create_intraday_price(instrument.symbol, price)
        
        for price_daily in data.price_daily_history:
            price_daily.market_id = market.id
            price_daily.instrument_id = instrument.id
            price_daily.imported_at = now
            self._session.add(price_daily)

        self.commit()


def gmt_time_now():
     return datetime.now().astimezone(ZoneInfo(key='GMT')) # in db its stored as gmt


__version__ = '0.1.0'       # semver MAJOR.MINOR.PATCH

# apply extension methods
from .shared import extensions
del(extensions)

# create default facade instance
from .facade import Facade
from .shared.container import Container

_container = Container()
def get_facade() -> Facade:
    from .facade import Facade
    return _container.resolve(Facade)

del(Container, Facade) # cleanup global namespace


from zoneinfo import ZoneInfo
from sqlalchemy import Column, Integer, Interval, String, ForeignKey, Table
from sqlalchemy.orm import relationship
from .base import Base


markets_instruments_association = Table('markets_instruments', Base.metadata,
    Column('market_id', Integer, ForeignKey('market.id')),
    Column('instrument_id', Integer, ForeignKey('instrument.id'))
)

class Market(Base):
    """Market where financial instruments are traded"""

    __tablename__ = 'market'
    id = Column(Integer, primary_key=True)
    #id = Column(UUID(as_uuid=True), primary_key=True, default=uuid_time_nextval)
    name = Column(String)
    description = Column(String)
    timezone_name = Column(String)

    instruments = relationship("Instrument", secondary=markets_instruments_association)



import pandas as pd
import yfinance as yf
from ..data.security_type import SecurityType


class Ticker:
    """Stock ticker is a report of the price of certain securities, updated continuously
    throughout the trading session by a particular stock market exchange."""
    
    def __init__(self, symbol: str):
        self.symbol = symbol
        self.history = pd.DataFrame()

    



import lagom
from typing import Any, TypeVar, Callable
from ..data.repository import Repository
from ..services.feed import Feed, NasdaqFeed


TAbstract = TypeVar("TAbstract")
T = TypeVar("T")
class Container:
    def __init__(self):
        self.container = lagom.Container()
        self.register_singleton(Repository)
        self.register_singleton_pair(Feed, NasdaqFeed)

    def register_singleton_pair(self, tabstract: TAbstract, tconcrete: T, override=False):
        if override and tabstract in self.container._registered_types:
            self.container._registered_types.pop(tabstract, None)
        self.container[tabstract] = lagom.Singleton(tconcrete)

    def register_singleton(self, type):
        self.container[type] = lagom.Singleton(type)
    
    def register_singleton_factory(self, type: T, factory: Callable[[], T]):
        self.container[type] = factory


    def resolve(self, type: T) -> T:
        result = self.container[type]
        if not result:
            raise Exception(f'unable to instantiate type {type}')
        return result



import pathlib, os
import string, random


def check_internet(host="8.8.8.8", port=53, timeout=3) -> bool:
    """
    tests if there is internet connection available
    taken from https://stackoverflow.com/questions/3764291/how-can-i-see-if-theres-an-available-and-active-network-connection-in-python
    Host: 8.8.8.8 (google-public-dns-a.google.com)
    OpenPort: 53/tcp
    Service: domain (DNS/TCP)
    """
    import socket
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except socket.error as ex:
        #print(ex)
        return False


def get_cache_folder_path():
    return pathlib.Path(os.environ.get('XDG_STATE_HOME') 
        or pathlib.Path.home().joinpath('.local', 'state')).joinpath('marketa', 'cache').resolve()


chars = string.ascii_letters + string.octdigits
def random_string_generator(str_size, allowed_chars = chars):
    return ''.join(random.choice(allowed_chars) for x in range(str_size))


def hookup_tqdm():
    import sys, marketa.shared.taskeeper as taskeeper, tqdm
    mod = sys.modules[__name__]
    # hookup tdqm with taskeeper
    def task_start(steps: int):
        mod.pbar = tqdm.tqdm(total = steps)
    taskeeper.on_start_callback = task_start
    def step_start(name: str):
        mod.pbar.set_description(name)
    taskeeper.on_step_start_callback = step_start
    def step_finish():
        mod.pbar.update()
    taskeeper.on_step_finish_callback = step_finish
    def task_finish(info: str):
        mod.pbar.set_description(info)
        mod.pbar.close()
    taskeeper.on_finish_callback = task_finish



from typing import Iterable
from marketa.data.instrument import Instrument
from marketa.data.price_daily import PriceDaily
from marketa.data.price_intraday import PriceIntraday


class InstrumentData:
    def __init__(self, 
                 instrument: Instrument, 
                 price_intraday: PriceIntraday, 
                 price_daily_history: Iterable[PriceDaily]):
        self.instrument = instrument
        self.price_intraday = price_intraday
        self.price_daily_history = price_daily_history


on_start_callback = lambda steps: ()
on_step_start_callback = lambda name: ()
on_step_finish_callback = lambda: ()
on_finish_callback = lambda info: ()

class Step:
    def __init__(self, name: str):
        on_step_start_callback(name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        on_step_finish_callback()


class Taskeeper:
    def __init__(self, steps: int):
        self.info: str = None
        on_start_callback(steps)

    def __enter__(self):
        return self
    
    def step(self, name: str) -> Step:
        return Step(name)

    def set_info(self, info:str):
        self.info = info

    def __exit__(self, exc_type, exc_value, traceback):
        on_finish_callback(self.info)
        

def taskeeper(steps:int):
    return Taskeeper(steps)





from .base import Base
from sqlalchemy import Column, DateTime, Integer, String, ForeignKey, Table, UniqueConstraint, Index

class Watchlist(Base):
   """Watchlist is a set of securities that an investor monitors 
      for potential trading or investing opportunities"""

   __tablename__ = 'watchlist'

   id = Column(Integer, primary_key=True)
   symbol = Column(String)
   updated_at = Column(DateTime)

   #name = Column(String)
   #UniqueConstraint('name', 'symbol', name='name_symbol')



from decimal import Decimal
from typing import Final, Iterable, List
import pandas as pd
import io, requests
from pandas import DataFrame
from purl import URL
from marketa.data.instrument_data import InstrumentData
from marketa.data.market import Market
from marketa.data.instrument import Instrument
from marketa.data.price_daily import PriceDaily
from marketa.data.price_intraday import PriceIntraday
from marketa.entities.ticker import Ticker
from ..data.security_type import SecurityType
from ..shared.exceptions import MarketaException
import yfinance as yf
from abc import ABC, abstractmethod


class Feed(ABC):

    @abstractmethod
    def load_market(self) -> Market: pass

    @abstractmethod
    def load_instruments(self) -> Iterable[Instrument]: pass

    @abstractmethod
    def load_instrument_data(self, symbol: str) -> InstrumentData: pass



class NasdaqFeed(Feed):
    """Responsible for importing data from nasdaq.com website pages that contain securities feeds.
       Feeds are produced by online stock screener. Stock screeners are tools that allow investors 
       and traders to sort through thousands of individual securities to find those that fit their own needs."""

    headers: Final = {'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.182 Safari/537.36'}

    def __init__(self):
        pass


    def load_market(self) -> Market:
        result = Market()
        result.name = 'NASDAQ'
        result.description = 'Nasdaq is a global electronic marketplace for buying and selling securities. Originally an acronym for "National Association of Securities Dealers Automated Quotations"—it was a subsidiary of the National Association of Securities Dealers (NASD), now known as the Financial Industry Regulatory Authority (FINRA). Nasdaq was created as a site where investors could trade securities on a computerized, speedy, and transparent system.1 It commenced operations on Feb. 8, 1971'
        # load any ticker to get exchange timezone
        t = yf.Ticker('AAAU')
        result.timezone_name = t.info['exchangeTimezoneName']
        return result


    def load_instruments(self) -> Iterable[Instrument]:
        data = DataFrame()
        data_traded = self.__fetch_nasdaq_traded_symbols()
        data_index = self.__fetch_nasdaq_index_symbols()

        # merge results into a single dataframe
        data = data.append(data_traded)
        data = data.append(data_index)
        data.reset_index(
            drop=True, # removes old incorrect integer row index and replaces it with updated one
            inplace=True
        )

        # convert into a typed collection
        result: List[Instrument] = []
        for idx, row in data.iterrows():
            instrument = Instrument()
            instrument.symbol = row['symbol']
            instrument.name = row['name']
            instrument.type = SecurityType.parse(row['type'])
            result.append(instrument)
        return result


    def load_instrument_data(self, symbol: str) -> InstrumentData:
        instrument = Instrument()
        price_intraday = PriceIntraday()
        price_daily_today = PriceDaily() # we will store potentially incomplete aggregate values available
        price_daily_history: List[PriceDaily] = []

        t = yf.Ticker(symbol)
        
        instrument.symbol = symbol
        instrument.name = t.info['longName']
        
        if 'quoteType' in dict(t.info).keys():
            instrument.type = SecurityType.parse(t.info['quoteType'])

        if 'longBusinessSummary' in dict(t.info).keys():
            instrument.description = t.info['longBusinessSummary']
        
        instrument.currency = t.info['currency']
        price_daily_today.volume = int(t.info['regularMarketVolume'])

        price_daily_today.open = Decimal(t.info['open'])
        price_daily_today.high = t.info['dayHigh']
        price_daily_today.low = t.info['dayLow']
        # close is ambiguous here, as it seems we don't have field indicating if market closed or not
        # and regularMarketPreviousClose may indicate close for previous day
        #if not t.info['tradeable']:
        #    price_daily_today.last_close = self.price # after market closes, last_close should equal to price
        #else:
        #    self.last_close = t.info['regularMarketPreviousClose']
        
        price_intraday.price = Decimal(t.info['regularMarketPrice'])
        price_intraday.bid = t.info['bid']
        price_intraday.bid_size = t.info['bidSize']
        price_intraday.ask = t.info['ask']
        price_intraday.ask_size = t.info['askSize']

        history:DataFrame = yf.download(symbol, period='max', progress=False)
        if len(history.index) == 0:
            raise Exception(f'no historical data for "{symbol}" symbol')

        for idx, row in history.iterrows():
            price_item = PriceDaily()
            price_item.open = row['Open']
            price_item.close = row['Close']
            price_item.high = row['High']
            price_item.low = row['Low']
            price_item.volume = row['Volume']
            price_item.date = row.name.to_pydatetime().date()
            price_daily_history.insert(0, price_item)

        price_daily_history.append(price_daily_today)
        return InstrumentData(instrument, price_intraday, price_daily_history)



    def __fetch_nasdaq_traded_symbols(self) -> DataFrame:
        url = URL('https://www.nasdaqtrader.com/dynamic/SymDir/nasdaqtraded.txt')
        csv_text = requests.get(url, headers = self.headers).text
        df: DataFrame = pd.read_csv(io.StringIO(csv_text), sep="|")
        
        # drop test rows
        df.drop_rows(column = 'Test Issue', predicate = lambda c: c == 'Y')

        df.drop(columns=[
            'Nasdaq Traded', 'Test Issue', 'Round Lot Size', 'Market Category', 
            'Listing Exchange', 'Financial Status', 'CQS Symbol', 'NASDAQ Symbol', 'NextShares'
        ], inplace=True)
        
        df.rename(columns={
            'Symbol': 'symbol',
            'Security Name': 'name',
            'ETF': 'etf',
        }, inplace=True)
        
        df['type'] = df['etf'].apply(lambda x: SecurityType.stock.name if x == 'N' else SecurityType.etf.name)
        df.drop(columns=['etf'], inplace=True)

        return df


    def __fetch_nasdaq_index_symbols(self) -> DataFrame:
        df = DataFrame()
        url = URL('https://api.nasdaq.com/api/screener/index')
        
        # we can only query 50 pages per request, otherwise it won't work
        offset = 0
        while True:
            url = URL('https://api.nasdaq.com/api/screener/index?offset=' + str(offset))
            json = requests.get(url, headers = self.headers).json()
            rows = json['data']['records']['data']['rows']
            for irow, row in enumerate(rows):
                df.append(pd.Series(dtype='object'), ignore_index=True) # insert empty row
                for col_name in row:
                    df.loc[irow + offset, col_name] = row[col_name]
            offset+=50
            if offset > json['data']['records']['totalrecords']:
                break
        
        df.drop(columns=['lastSalePrice', 'netChange', 'percentageChange', 'deltaIndicator'], inplace=True)
        df.rename(columns={ 'companyName': 'name' }, inplace=True)
        df = df.assign(type=SecurityType.index.name) # adds 'type' attribute with 'index' value to all rows
        if 'symbol' not in df.columns:
            raise MarketaException('nasdaq index feed is missing "symbol" column')
        return df



from sqlalchemy import Column, Integer, ForeignKey, Numeric, DateTime
from .base import Base


class PriceIntraday(Base):
    """Intraday price of an instrument traded on a particular market at particular time and date"""

    __tablename__ = 'prices_intraday'
    id = Column(Integer, primary_key=True)
    #id = Column(UUID(as_uuid=True), primary_key=True, default=uuid_time_nextval)

    market_id = Column(Integer, ForeignKey('market.id'))
    instrument_id = Column(Integer, ForeignKey('instrument.id'))
    
    # market price of an instrument
    price = Column(Numeric)
    
    # orderbook aggregate data
    bid = Column(Numeric)
    bid_size = Column(Integer)
    ask = Column(Numeric)
    ask_size = Column(Integer)

    datetime = Column(DateTime)      # when instrument had that price
    imported_at = Column(DateTime)   # when price was imported




from enum import Enum

from marketa.shared.exceptions import MarketaException

class SecurityType(Enum):
    stock = 0   # common stock
    mutual = 1  # mutual fund
    etf = 2     # exchange traded fund
    index = 3   # index fund

    def parse(s: str):
        if s.upper() == 'ETF': return SecurityType.etf
        if s.upper() == 'INDEX': return SecurityType.index
        if s.upper() == 'STOCK': return SecurityType.stock
        raise MarketaException(f'unknown security type {s}')



from sqlalchemy import Column, DateTime, Enum, Integer, Numeric, ForeignKey, Date
from .base import Base


class PriceDaily(Base):
    """Daily report of a price of an instrument traded on a particular market at particular date"""

    __tablename__ = 'prices_daily'
    id = Column(Integer, primary_key=True)
    market_id = Column(Integer, ForeignKey('market.id'))
    instrument_id = Column(Integer, ForeignKey('instrument.id'))
    #id = Column(UUID(as_uuid=True), primary_key=True, default=uuid_time_nextval)

    
    open = Column(Numeric)
    close = Column(Numeric)
    high = Column(Numeric)
    low = Column(Numeric)

    volume = Column(Integer)

    date = Column(Date)                # day when instrument had that price
    imported_at = Column(DateTime)   # when price was imported


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['marketa',
 'marketa.data',
 'marketa.entities',
 'marketa.services',
 'marketa.shared']

package_data = \
{'': ['*']}

install_requires = \
['PyQt5>=5.15.6,<6.0.0',
 'PyYAML>=6.0,<7.0',
 'SQLAlchemy-Utils>=0.38.2,<0.39.0',
 'SQLAlchemy>=1.4.31,<2.0.0',
 'daemonize>=2.5.0,<3.0.0',
 'diskcache>=5.2.1,<6.0.0',
 'docopt>=0.6.2,<0.7.0',
 'exitstatus>=2.2.0,<3.0.0',
 'financedatabase>=1.0.0,<2.0.0',
 'humanize>=3.13.1,<4.0.0',
 'lagom>=1.7.0,<2.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'mplfinance>=0.12.8-beta.9,<0.13.0',
 'numpy>=1.20.2,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'purl>=1.5,<2.0',
 'sequential-uuids>=1.0.0,<2.0.0',
 'tqdm>=4.62.3,<5.0.0',
 'yfinance>=0.1.63,<0.2.0']

entry_points = \
{'console_scripts': ['linter = poetry:lint']}

setup_kwargs = {
    'name': 'marketa',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'illiak',
    'author_email': 'ilichpost@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import uuid
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy import Column, Integer, String, Enum, Numeric, ForeignKey, Table
from sqlalchemy.orm import relationship
from sequential_uuids.generators import uuid_time_nextval
from .security_type import SecurityType
from .base import Base
from .price_daily import PriceDaily
from .price_intraday import PriceIntraday


class Instrument(Base):
    """Financial instruments are assets that can be traded, or they
       can also be seen as packages of capital that may be traded."""
    
    __tablename__ = 'instrument'

    id = Column(Integer, primary_key=True)
    #id = Column(UUID(as_uuid=True), primary_key=True, default=uuid_time_nextval)
    name = Column(String)
    symbol = Column(String)
    type = Column(Enum(SecurityType))
    description = Column(String)
    currency = Column(String)
    
    prices_daily = relationship(PriceDaily)
    prices_intraday = relationship(PriceIntraday)



